"""Defines constants"""

G = 6.67384e-11
RSOL = 6.955e8
RJUP = 7.1492e7
RJUP_POL = 6.9911e7 # Polar Radius
PI = 3.14159265359
MSOL = 1.98847542e+30
MJUP = 1.898e27
AU = 1.49597871e+11
KBOLTZ = 1.380648813e-23
RGP = 8.31446
PLANCK = 6.62606957e-34
C_LUM = 299792458
SIG_SB = 5.670367e-08
N_A = 6.022140857e+23
tiny=1.e-13
OneMintiny=1.-tiny

ktable_long_name_attributes={
    'p': 'Pressure grid',
    't': 'Temperature grid',
    'x': 'Volume mixing ratio grid',
    'bin_centers': 'Centers of the wavenumber bins',
    'bin_edges': 'Separations between the wavenumber bins',
    'weights': 'Weights used in the g-space quadrature',
    'samples': 'Abscissas used to sample the k-coefficients in g-space',
    'mol_name': 'Name of the species described',
    'mol_mass': 'Mass of the species',
    'kcoeff': 'Table of the k-coefficients with axes (pressure, temperature, wavenumber, g space)',
    'xsecarr': 'Table of cross sections with axes (pressure, temperature, wavenumber)',
    'method': 'Name of the method used to sample g-space',
    'ngauss': 'Number of points used to sample the g-space',
    'temperature_grid_type': 'Whether the temperature grid is "regular" (same temperatures for all pressures) or "pressure-dependent"',
    'wnrange': 'Wavenumber range covered',
    'wlrange': 'Wavelength range covered',
    'DOI': 'Data object identifier linked to the data',
    'Date_ID':'Date at which the table has been created, along with the version of exo_k',
    'key_iso_ll': 'Isotopologue identifier'
}

nemesis_hitran_id_numbers={
'H2O': 1,
'CO': 5,
'CO2': 2,
'O2': 7,
'O3': 3,
'N2O': 4,
'AlCl': 84,
'AlF': 85,
'Al': 86,
'AlO': 83,
'AsH3': 41,
'BeH': 87,
'C2': 88,
'C2H2': 26,
'C2H4': 32,
'CaF': 89,
'CaH': 90,
'CaO': 92,
'CH': 93,
'CH3': 94,
'CH3Cl': 24,
'CH3F': 95,
'CH4': 6,
'CN': 96,
'CP': 97,
'CrH': 98,
'CS': 54,
'FeH': 82,
'H2': 39,
'H2CO': 20,
'H2O2': 25,
'H2S': 36,
'H3^+': 81,
'HBr': 16,
'HCl': 15,
'HCN': 23,
'HD+': 99,
'HeH+': 100,
'HF': 14,
'HI': 17,
'HNO3': 12,
'K': 61,
'KCl': 101,
'KF': 102,
'LiCl': 103,
'LiF': 104,
'LiH': 105,
'LiH+': 106,
'MgF': 107,
'MgH': 108,
'MgO': 109,
'Na': 60,
'NaCl': 110,
'NaF': 111,
'NaH': 112,
'NH': 113,
'NH3': 11,
'NO': 8,
'NO2': 10,
'NS': 114,
'OH': 13,
'OH+': 115,
'P2H2': 116,
'PH': 118,
'PH3': 28,
'PN': 119,
'PO': 120,
'PS': 121,
'ScH': 122,
'SH': 123,
'SiH': 124,
'SiH2': 125,
'SiH4': 126,
'SiO': 127,
'SiO2': 128,
'SiS': 129,
'SO2': 9,
'SO3': 74,
'TiH': 130,
'TiO': 62,
'VO': 63
}
